/* Generated by Cython 0.29.34 */

#ifndef __PYX_HAVE__libzim
#define __PYX_HAVE__libzim

#include "Python.h"

#ifndef __PYX_HAVE_API__libzim

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C bool obj_has_attribute(PyObject *, std::string);
__PYX_EXTERN_C std::string string_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C wrapper::Blob blob_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C zim::writer::ContentProvider *contentprovider_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C zim::writer::IndexData *indexdata_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C bool bool_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C uint64_t uint64_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C uint32_t uint32_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C zim::writer::IndexData::GeoPosition geoposition_cy_call_fct(PyObject *, std::string, std::string *);
__PYX_EXTERN_C std::map<zim::writer::HintKeys,uint64_t>  convertToCppHints(PyObject *);
__PYX_EXTERN_C std::map<zim::writer::HintKeys,uint64_t>  hints_cy_call_fct(PyObject *, std::string, std::string *);

#endif /* !__PYX_HAVE_API__libzim */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initlibzim(void);
#else
PyMODINIT_FUNC PyInit_libzim(void);
#endif

#endif /* !__PYX_HAVE__libzim */
